from typing import Optional, Union

import discord
from cpl_discord.container import Role, CategoryChannel
from cpl_discord.service import DiscordBotServiceABC
from discord import Guild, Member
from discord.abc import GuildChannel

from bot_data.abc.realm_repository_abc import RealmRepositoryABC
from bot_data.model.server import Server


class RealmUtils:
    def __init__(self, bot: DiscordBotServiceABC, realms: RealmRepositoryABC):
        self._bot = bot
        self._realms = realms

    @staticmethod
    def find_role_by_name(guild: Guild, name: str) -> Optional[Role]:
        for role in guild.roles:
            if role.name == name:
                return Role(role)
        return None

    @staticmethod
    def get_role_by_name(guild: Guild, name: str) -> Optional[Role]:
        for role in guild.roles:
            if role.name == name:
                return Role(role)
        raise ValueError(f"Role {name} not found")

    @staticmethod
    def find_category_by_name(guild: Guild, name: str) -> Optional[CategoryChannel]:
        for category in guild.categories:
            if category.name == name:
                return CategoryChannel(category)
        return None

    @staticmethod
    def get_category_by_name(guild: Guild, name: str) -> Optional[CategoryChannel]:
        for category in guild.categories:
            if category.name == name:
                return CategoryChannel(category)
        raise ValueError(f"Category {name} not found")

    @staticmethod
    def check_realm_role(guild: Guild, role: Role) -> bool:
        return role in guild.roles

    @staticmethod
    async def create_category(guild: Guild, name: str, role: Role, leader: Optional[Member]) -> CategoryChannel:
        overwrites = {
            guild.default_role: discord.PermissionOverwrite(view_channel=False, connect=False),
            role: discord.PermissionOverwrite(view_channel=True, connect=True),
        }
        if leader is not None:
            overwrites[leader] = discord.PermissionOverwrite(
                view_channel=True,
                connect=True,
                manage_channels=True,
                manage_permissions=True,
                create_instant_invite=True,
            )

            if leader.get_role(role.id) is None:
                await leader.add_roles(role)

        return await guild.create_category(name, overwrites=overwrites)

    @staticmethod
    async def assign_moderator(category: Union[CategoryChannel, GuildChannel], member: Member):
        await category.set_permissions(
            member,
            view_channel=True,
            connect=True,
            manage_channels=True,
            manage_permissions=True,
            create_instant_invite=True,
        )

    @staticmethod
    async def remove_moderator(category: Union[CategoryChannel, GuildChannel], member: Member):
        await category.set_permissions(
            member,
            view_channel=True,
            connect=True,
            manage_channels=False,
            manage_permissions=False,
            create_instant_invite=False,
        )

    def is_realm_moderator(self, member_id: int, realm_id: int) -> bool:
        return (
            self._realms.get_realm_moderators_by_realm_id(realm_id)
            .select(lambda x: x.user.discord_id)
            .contains(member_id)
        )

    async def check_integrity(self, server: Server, guild: Guild):
        for realm in self._realms.get_realms_by_server_id(server.id):
            category = self.find_category_by_name(guild, realm.name)
            if category is None:
                await self.create_category(guild, realm.name, realm.role, None)

            role = self.find_role_by_name(guild, realm.name)
            if role is None:
                await guild.create_role(name=realm.name)

            role = self.get_role_by_name(guild, realm.name)
            category = self.get_category_by_name(guild, realm.name)
            for mod in self._realms.get_realm_moderators_by_realm_id(realm.id):
                member = guild.get_member(mod.user.discord_id)
                if member is None:
                    continue

                if member.get_role(role.id) is None:
                    await member.add_roles(role)

                if member not in category.overwrites:
                    await self.assign_moderator(category, member)

            mods = self._realms.get_realm_moderators_by_realm_id(realm.id)
            mod_user_ids = mods.select(lambda x: x.user.discord_id)
            for mod in category.overwrites:
                if isinstance(mod, discord.Role):
                    continue

                if mod.id not in mod_user_ids:
                    await self.remove_moderator(category, mod)
